C****************************************************************************
C If you use this code in any publication, please cite Verde et al. (2003),
C if you use the accompanying data files, please also reference 
C Hinshaw et al. (2003) and Kogut et al. (2003). 
C****************************************************************************
C This file contains a set of Fortran subroutines that compute likelihoods
C for a set of theoretical cl's:
C    common_t_likelihood      - Fills the temperature data common block.
C                               Call this routine once before calling
C                               compute_mapt_likelihood
C    common_te_likelihood     - Fills the TE data common block.
C                               Call this routine once before calling
C                               compute_mapte_likelihood
C    common_all_likelihood    - Calls common_t_likelihood and 
C                               common_te_likelihood.
C    compute_mapt_likelihood  - Computes temperature data likelihoods.
C    compute_mapt_fisher      - Returns the TT Fisher matrix.
C    compute_mapte_likelihood - Computes TE data likelihoods.
C    compute_mapte_fisher     - Returns the TE Fisher matrix.
C
C The methods used here are described in the following references:
C    Verde, L., et.al. 2003, ApJ, submitted.
C    Bond, J.R., Jaffe, A.H., and Knox, L., 2002, ApJ, 533
C If you use this software in preparing results in any publication, please
C reference Verde etal, 2003.
C
C This software should successfully compile using a 
C Fortran 90 compiler.  Interface blocks describing the calling sequences
C for the subroutines is supplied in "compute_likelihood.fh".
C
C You should:
C 1)  At the start of the program, fill the common blocks. (use common_[...] routines)
C 2)  Calculate the theory cls. (use cmbfast or your favourite alternative)
C 3)  call the likelihood functions (use the compute_[....] routines)
C  Good luck!
C
C ===========================================================================
C   "common_t_likelihood" fills the common block used by
C   "compute_mapt_likelihood"---this is the initialization subroutine.
C
C   Inputs:
C      clFile  - The name of the file containing the cl data.
C      offDiag - The name of the file containing the off-diagonal terms.
C
C   Outputs:
C      stat - A status code:  0=success.
C
C   Common blocks:
C        tt_data - This common block is used to pass/store information
C                  that does not change between calls.
C
C   Written by Licia Verde and Hiranya Peiris, Princeton University,
C      December 2002.
C ===========================================================================
        SUBROUTINE common_t_likelihood (clFile, offDiag, stat)
C
        IMPLICIT NONE
C
	integer tarrsiz
	parameter (tarrsiz=900)
C
C                INPUT
C                
        character (*) clFile, offDiag
C
C                OUTPUT
C
        integer stat
C
C		 DATA/COMMON BLOCK
C
        real*8 cl_data(tarrsiz), neff(tarrsiz), fskyeff(tarrsiz)
        real*8 r_off_diag(tarrsiz,tarrsiz), off_diag(tarrsiz,tarrsiz)
        common /tt_data/ cl_data,neff,fskyeff,r_off_diag,off_diag
C
C                DATA/INTERNAL
C
        integer idum, l, ll
C ---------------------------------------------------------------------------
C		Read the CL data.
C
        open (11, file=clFile, status='old', IOStat=stat)
	if (stat .NE. 0) RETURN
	rewind(11)
	do l = 2, tarrsiz
	  read (11, *) idum, cl_data(l), neff(l), fskyeff(l)
	end do
	close (11)
C
C		Read in off diag terms:		      
C		In the *covariance* matrix there are 2 type of terms 
C		one that does scale with the clth (due to the mask) and 
C		one that does not (due to beam and point sources marginalization).
C 		In the curvature matrix (what we use here) all the off diagonal 
C		terms end up scaling with the clth but the 2 contributions scale
C		in different ways see paper for details
C 		thus here we read them in separately
C
	open(11,file=offDiag, status='old', IOStat=stat)
	if (stat .NE. 0) RETURN
	rewind(11)
	do  l = 2,tarrsiz	
	   do ll= l+1,tarrsiz
	      read(11,*) idum,idum,off_diag(l,ll),r_off_diag(l,ll)
	      r_off_diag(ll,l)=r_off_diag(l,ll)
	      off_diag(ll,l)=off_diag(l,ll)
	   end do
	end do
	close(11)
C
        RETURN
C ---------------------------------------------------------------------------
        END
C ===========================================================================
C   "common_te_likelihood" fills the common block used by
C   "compute_mapte_likelihood"---this is the initialization subroutine.
C
C   Inputs:
C      clFile  - The name of the file containing the cl data.
C      offDiag - The name of the file containing the off-diagonal terms.
C
C   Outputs:
C      stat - A status code:  0=success.
C
C   Common blocks:
C        te_dat  - This common block is used to pass/store information
C                  that does not change between calls.
C
C   Written by Licia Verde and Hiranya Peiris, Princeton University,
C      December 2002.
C ===========================================================================
        SUBROUTINE common_te_likelihood (clFile, offDiag, stat)
C
        IMPLICIT NONE
C
	integer tearrsiz
	parameter (tearrsiz=512)
C
	integer d_mapte_ltot
	parameter (d_mapte_ltot=512)	! max l for te data
C
C
C                INPUT
C                
        character (*) clFile, offDiag
C
C                OUTPUT
C
        integer stat
C
C		 DATA/COMMON BLOCK
C
        real*8 te_data(tearrsiz),ntt(tearrsiz),nee(tearrsiz)
	real*8 te_off_diag(tearrsiz,tearrsiz)
        real*8 cltt_data(tearrsiz)
        common/te_dat/ te_data,ntt,nee,te_off_diag,cltt_data
C
C                DATA/INTERNAL
C
        integer idum, l, ll
C ---------------------------------------------------------------------------
C		Read the CL data.
C
        open (11, file=clFile, status='old', IOStat=stat)
	if (stat .NE. 0) RETURN
	rewind(11)
	do  l = 2, d_mapte_ltot
	   read(11,*) idum,te_data(l),cltt_data(l),ntt(l),nee(l)
	end do
	close(11)
C
C		Read in off diag terms.
C
	open(11,file=offDiag, status='old', IOStat=stat)
	if (stat .NE. 0) RETURN
	rewind(11)
	do  l = 2,d_mapte_ltot	
	   do ll= l+1,d_mapte_ltot	
	      read(11,*) idum,idum,te_off_diag(l,ll)
	      te_off_diag(ll,l)=te_off_diag(l,ll)
	   end do
	end do
	close(11)
C
        RETURN
C ---------------------------------------------------------------------------
        END
C ===========================================================================
C   "common_all_likelihood" encapsulates "common_t_likelihood" and
C   "common_te_likelihood" into a single call.
C
C   Inputs:
C      TclFile   - The name of the file containing the temp. cl data.
C      ToffDiag  - The name of the file containing the temp. off-diagonal terms.
C      TEclFile  - The name of the file containing the TE cl data.
C      TEoffDiag - The name of the file containing the TE off-diagonal terms.
C
C   Outputs:
C      stat - A status code:  0=success.
C
C   Written by Licia Verde and Hiranya Peiris, Princeton University,
C      December 2002.
C ===========================================================================
        SUBROUTINE common_all_likelihood (TclFile, ToffDiag,
     1                                    TEclFile, TEoffDiag, stat)
C
        IMPLICIT NONE
C
C                INPUT
C                
        character (*) TclFile, ToffDiag, TEclFile, TEoffDiag
C
C                OUTPUT
C
        integer stat
C ---------------------------------------------------------------------------
        Call common_t_likelihood (TclFile, ToffDiag, stat)
	If (stat .NE. 0) Return
        Call common_te_likelihood (TEclFile, TEoffDiag, stat)
	Return
C ---------------------------------------------------------------------------
        END
C ===========================================================================
C   "compute_mapt_likelihood" computes the likelihood for temperature data
C    using the form for the likelihood as in Verde et al 2003 sec. 2.1 
C
C    IF THE C_l AMPLITUDE IS TOO LOW, THE APROXIMATION MADE IN COMPUTING THE 
C    CURVATURE MATRIX AND THE LIKELIHOOD BREAKS DOWN.  
C    l*(l+1)/2pi * C_l AT THE FIRST PEAK MUST BE >~ 5000 
C
C    There are 2 contributions to the  off diagonal terms of the fisher matrix: 
C    this is because in the covariance matrix the terms due to marginalization over 
C    point sources and beam uncertainties depend on the power spectrum that's out 
C    there in the sky while the coupling due to the mask depend on the Cl_theory 
C    (i.e. changes as the cosmological parameters change in exploring the 
C    likelihood surface) see Verde et al. 2003 , Hinshaw et al 2003.
C
C    WARNING: a 2% bias around the first peak results from incorrect scaling 
C    of the off diagonal terms.    
C
C   Inputs:
C      clth - An array of cl's describing the theory being tested.
C
C   Outputs:
C      l0   - The natural log of the likelihood for l >= 2.
C
C   Common blocks:
C        tt_data - This common block is used to pass/store information
C                  that does not change between calls.
C
C   Written by Licia Verde and Hiranya Peiris, Princeton University,
C      December 2002.
C ===========================================================================
        SUBROUTINE compute_mapt_likelihood (clth, l0)
C
        IMPLICIT NONE
C
	integer tarrsiz
	parameter (tarrsiz=900)
C
C                INPUT: clt_theory
C                
        real*8 clth(:)
C
C                OUTPUT: log(likelihood)
C
        real*8 l0
C
C		 DATA/COMMON BLOCK
C
        real*8 cl_data(tarrsiz), neff(tarrsiz), fskyeff(tarrsiz)
        real*8 r_off_diag(tarrsiz,tarrsiz), off_diag(tarrsiz,tarrsiz)
        common /tt_data/ cl_data,neff,fskyeff,r_off_diag,off_diag
C
C                DATA/INTERNAL
C
        integer l,ll,idum,i,j,info, d_mapt_lmax, d_mapt_lmin
        integer larr(tarrsiz)
        real*8 loglike, dum, error, newlike
        real*8 dloglike, Fisher(tarrsiz,tarrsiz), Fdiag(tarrsiz)
	real*8 z(tarrsiz), zbar(tarrsiz)
        real*8 off_log_curv(tarrsiz,tarrsiz)
C ---------------------------------------------------------------------------
        loglike   = 0.0d0
C
        d_mapt_lmax = tarrsiz  ! lmax
	d_mapt_lmin =      2   ! lmin
C        
C prepare to compute the offset longormal likelihood as in Bond Jaffe Knox.
C with the difference that the transformation we do on the curvature
C matrix is using clth not cltdata
C this is closer to the equal variance approx (see Bond Jaffe Knox again)  
C form more details see Verde et al .2003 
C
C here Fdiag is the diagonal term of the covariance matrix 
C Fisher denotes the curvature matrix.
C
        do l = 2, d_mapt_lmax        
           larr(l)  = l
           Fdiag(l) = 2.*(Clth(l)+neff(l))**2
     1              / ((2.0d0*dble(larr(l))
     2              +1.0d0)*fskyeff(l)*fskyeff(l))
           Fisher(l,l)=1.d0/Fdiag(l)
           z(l)=dlog(cl_data(l)+neff(l))           
           zbar(l)=dlog(clth(l)+neff(l))
        end do

        do l = 2,d_mapt_lmax        
           do ll=2,d_mapt_lmax
              if(l.ne.ll) then
C
C here the two contributions to the off diagonal terms to the curvature matrix 
C are treated separately
C
C see Verde et.al. 2003 for details.
C
                 Fisher(l,ll)=r_off_diag(l,ll)/dsqrt(Fdiag(l)*Fdiag(ll))
     1                       +off_diag(l,ll)/(Fdiag(l)*Fdiag(ll))
              endif 
              off_log_curv(l,ll)=(clth(l)+neff(l))
     2                          *Fisher(l,ll)*(clth(ll)+neff(ll))
           end do
        end do
C
        do l=d_mapt_lmin,d_mapt_lmax        
           do ll=d_mapt_lmin, d_mapt_lmax
C
C to correct for residual 0.5% bias around the peak
C see Verde et.al. 2003 for more details
C
C this is an interpolation between Bond Knox Jaffe and Gaussian Likelihood.
C works extremely well on sims (again see paper for details)
C
              dloglike=2./3.*(z(l)-zbar(l))*off_log_curv(l,ll)
     1                      *(z(ll)-zbar(ll))+1./3.*(clth(l)-cl_data(l))
     2                      *Fisher(l,ll)*(clth(ll)-cl_data(ll)) 
              loglike=loglike+dloglike
           end do
        end do
C
        l0=-loglike/2.d0
C
C        write(*,*) ' MAP_T log(l0)=',l0,'chisq=',loglike
C
        RETURN
C ---------------------------------------------------------------------------
        END
C ===========================================================================
C   "compute_mapt_fisher" returns the Fisher matrix for temperature data
C    using the form for the likelihood as in Verde et al 2003 sec. 2.1 
C
C    IF THE C_l AMPLITUDE IS TOO LOW, THE APROXIMATION MADE IN COMPUTING THE 
C    CURVATURE MATRIX AND THE LIKELIHOOD BREAKS DOWN.  
C    l*(l+1)/2pi * C_l AT THE FIRST PEAK MUST BE >~ 5000 
C
C    There are 2 contributions to the  off diagonal terms of the fisher matrix: 
C    this is because in the covariance matrix the terms due to marginalization over 
C    point sources and beam uncertainties depend on the power spectrum that's out 
C    there in the sky while the coupling due to the mask depend on the Cl_theory 
C    (i.e. changes as the cosmological parameters change in exploring the 
C    likelihood surface) see Verde et al. 2003 , Hinshaw et al 2003.
C
C    WARNING: a 2% bias around the first peak results from incorrect scaling 
C    of the off diagonal terms.    
C
C   Inputs:
C      clth - An array of cl's describing the theory being tested.
C
C   Outputs:
C      Fisher - The Fisher matrix.  Should be dimensioned at least 900x900.
C
C   Common blocks:
C        tt_data - This common block is used to pass/store information
C                  that does not change between calls.
C
C   Adapted from compute_mapt_likelihood by Licia Verde and Hiranya Peiris,
C      Princeton University, December 2002.
C ===========================================================================
        SUBROUTINE compute_mapt_fisher (clth, Fisher)
C
        IMPLICIT NONE
C
	integer tarrsiz
	parameter (tarrsiz=900)
C
C                INPUT: clt_theory
C                
        real*8 clth(:)
C
C                OUTPUT: log(likelihood)
C
        real*8 Fisher(:,:)
C
C		 DATA/COMMON BLOCK
C
        real*8 cl_data(tarrsiz), neff(tarrsiz), fskyeff(tarrsiz)
        real*8 r_off_diag(tarrsiz,tarrsiz), off_diag(tarrsiz,tarrsiz)
        common /tt_data/ cl_data,neff,fskyeff,r_off_diag,off_diag
C
C                DATA/INTERNAL
C
        integer l,ll,idum,i,j,info, d_mapt_lmax, d_mapt_lmin
        integer larr(tarrsiz)
        real*8 Fdiag(tarrsiz)
	real*8 z(tarrsiz), zbar(tarrsiz)
        real*8 off_log_curv(tarrsiz,tarrsiz)
C ---------------------------------------------------------------------------
        d_mapt_lmax = tarrsiz  ! lmax
	d_mapt_lmin =      2   ! lmin
C        
C prepare to compute the offset longormal likelihood as in Bond Jaffe Knox.
C with the difference that the transformation we do on the curvature
C matrix is using clth not cltdata
C this is closer to the equal variance approx (see Bond Jaffe Knox again)  
C form more details see Verde et al .2003 
C
C here Fdiag is the diagonal term of the covariance matrix 
C
        do l = 2, d_mapt_lmax        
           larr(l)  = l
           Fdiag(l) = 2.*(Clth(l)+neff(l))**2
     1              / ((2.0d0*dble(larr(l))
     2              +1.0d0)*fskyeff(l)*fskyeff(l))
           Fisher(l,l)=1.d0/Fdiag(l)
           z(l)=dlog(cl_data(l)+neff(l))           
           zbar(l)=dlog(clth(l)+neff(l))
        end do

        do l = 2,d_mapt_lmax        
           do ll=2,d_mapt_lmax
              if(l.ne.ll) then
C
C here the two contributions to the off diagonal terms to the curvature matrix 
C are treated separately
C
C see Verde et.al. 2003 for details.
C
                 Fisher(l,ll)=r_off_diag(l,ll)/dsqrt(Fdiag(l)*Fdiag(ll))
     1                       +off_diag(l,ll)/(Fdiag(l)*Fdiag(ll))
              endif 
              off_log_curv(l,ll)=(clth(l)+neff(l))
     2                          *Fisher(l,ll)*(clth(ll)+neff(ll))
           end do
        end do
C
        RETURN
C ---------------------------------------------------------------------------
        END
C ===========================================================================
C "compute_mapte_likelihood" computes the likelihood for TE data.
C
C IF THE C_l AMPLITUDE IS TOO LOW, THE APROXIMATION MADE IN COMPUTING THE 
C CURVATURE MATRIX AND THE LIKELIHOOD BREAKS DOWN.  
C l*(l+1)/2pi * C_l AT THE FIRST PEAK MUST BE >~ 5000 
C
C Uses the expression for the likelihood  as in Kogut et al 2003 (ApJ in press) 
C and Verde et al 2003.
C
C The Curvature matrix has been calibrated from Monte Carlo simulations.  
C fsky is set to 0.85, if not using the P2 cut (i.e. if not using the cl 
C we provided you with) you should change fsky accordingly.
C
C   Inputs:
C      cltt - An array of temperature cl's describing the theory
C             being tested.
C      clte - An array of TE cl's describing the theory
C             being tested.
C      clee - An array of EE cl's describing the theory
C             being tested.
C
C   Outputs:
C      l0   - The natural log of the likelihood for all l.
C
C   Common blocks:
C        te_dat - This common block is used to pass/store information
C                 that does not change between calls.
C
C   Written by Licia Verde and Hiranya Peiris, Princeton University,
C      December 2002.
C ===========================================================================
        SUBROUTINE compute_mapte_likelihood (cltt, clte, clee, l0)
C
        IMPLICIT NONE
C
	integer tearrsiz
	parameter (tearrsiz=512)
C
	integer d_mapte_ltot
	parameter (d_mapte_ltot=450)	! max l for te data
C
C
C		INPUT: clte_theory
C                
        real*8 cltt(:),clte(:),clee(:)
C
C		OUTPUT: log(likelihood)
C
        real*8 l0
C
C		 DATA/COMMON BLOCK
C
        real*8 te_data(tearrsiz),ntt(tearrsiz),nee(tearrsiz)
	real*8 te_off_diag(tearrsiz,tearrsiz)
        real*8 cltt_data(tearrsiz)
        common/te_dat/ te_data,ntt,nee,te_off_diag,cltt_data
C
C		DATA/INTERNAL
C
        integer il,ill,l,larr(tearrsiz)
        real*8 fsky,delta_loglike,loglike,dum
        real*8 dloglike,Fisher(tearrsiz,tearrsiz),Fdiag(tearrsiz)
C ---------------------------------------------------------------------------
        fsky=0.85d0
C
        loglike=0.0d0
C
C Fdiag is the diagonal element of the covariance matrix
C Fisher denotes the curvature matrix.
C
        do il = 2,d_mapte_ltot        
           larr(il)=il
           Fdiag(il)=((cltt(il)+ntt(il))*(clee(il)+nee(il))
     1              +clte(il)*clte(il))
     2              /((2.0d0*dble(larr(il))+1.0d0)*fsky*fsky/1.14d0)
C
C this correction factor (1.14) has been obtained from a calibration of
C the covariance matrix on monte carlo sims (see papers for more details)
C
           Fisher(il,il)=1.d0/Fdiag(il)
        end do
C
C include mask effect on off diagonal terms in the curvature matrix
C
        do il = 2,d_mapte_ltot
           do ill=2,d_mapte_ltot
              if(il.ne.ill) then
                 Fisher(il,ill)=te_off_diag(il,ill)/
     1                          dsqrt(Fdiag(il)*Fdiag(ill))
              endif 
           end do
        end do

        do il = 2,d_mapte_ltot
           do ill = 2, d_mapte_ltot
              delta_loglike = (clte(il)-te_data(il))*Fisher(il,ill)*
     1                        (clte(ill)-te_data(ill))
              loglike=loglike+delta_loglike
           end do          
        end do
C
        l0=-loglike/2.d0
C
C        write(*,*) ' MAP_TE log(l0)=',l0,'chisq=',loglike
C
        RETURN
C ---------------------------------------------------------------------------
        END
C ===========================================================================
C "compute_mapte_fisher" returns the Fisher matrix for TE data.
C
C IF THE C_l AMPLITUDE IS TOO LOW, THE APROXIMATION MADE IN COMPUTING THE 
C CURVATURE MATRIX AND THE LIKELIHOOD BREAKS DOWN.  
C l*(l+1)/2pi * C_l AT THE FIRST PEAK MUST BE >~ 5000 
C
C The Curvature matrix has been calibrated from Monte Carlo simulations.  
C fsky is set to 0.85, if not using the P2 cut  (i.e. if not using the cl
C we provided you with) you should change fsky accordingly.
C
C   Inputs:
C      cltt - An array of temperature cl's describing the theory
C             being tested.
C      clte - An array of TE cl's describing the theory
C             being tested.
C      clee - An array of EE cl's describing the theory
C             being tested.
C
C   Outputs:
C      Fisher - The Fisher matrix.  Should be dimensioned at least 512x512.
C
C   Common blocks:
C        te_dat - This common block is used to pass/store information
C                 that does not change between calls.
C
C   Adapted from compute_mapte_likelihood by Licia Verde and Hiranya Peiris,
C      Princeton University, December 2002.
C ===========================================================================
        SUBROUTINE compute_mapte_fisher (cltt, clte, clee, Fisher)
C
        IMPLICIT NONE
C
	integer tearrsiz
	parameter (tearrsiz=512)
C
	integer d_mapte_ltot
	parameter (d_mapte_ltot=450)	! max l for te data
C
C
C		INPUT: clte_theory
C                
        real*8 cltt(:),clte(:),clee(:)
C
C		OUTPUT: log(likelihood)
C
        real*8 Fisher(:,:)
Cg
C		 DATA/COMMON BLOCK
C
        real*8 te_data(tearrsiz),ntt(tearrsiz),nee(tearrsiz)
	real*8 te_off_diag(tearrsiz,tearrsiz)
        real*8 cltt_data(tearrsiz)
        common/te_dat/ te_data,ntt,nee,te_off_diag,cltt_data
C
C		DATA/INTERNAL
C
        integer il,ill,l,larr(tearrsiz)
        real*8 fsky
        real*8 Fdiag(tearrsiz)
C ---------------------------------------------------------------------------
        fsky=0.85d0
C
C Fdiag is the diagonal element of the covariance matrix
C
        do il = 2,d_mapte_ltot        
           larr(il)=il
           Fdiag(il)=((cltt(il)+ntt(il))*(clee(il)+nee(il))
     1              +clte(il)*clte(il))
     2              /((2.0d0*dble(larr(il))+1.0d0)*fsky*fsky/1.14d0)
C
C this correction factor (1.14) has been obtained from a calibration of
C the covariance matrix on monte carlo sims (see papers for more details)
C
           Fisher(il,il)=1.d0/Fdiag(il)
        end do
C
C include mask effect on off diagonal terms in the curvature matrix
C
        do il = 2,d_mapte_ltot
           do ill=2,d_mapte_ltot
              if(il.ne.ill) then
                 Fisher(il,ill)=te_off_diag(il,ill)/
     1                        dsqrt(Fdiag(il)*Fdiag(ill))
              endif 
           end do
        end do
C
        RETURN
C ---------------------------------------------------------------------------
        END
